/*
 This file is part of the digitalSTROM ds485-client library.

 Copyright (C) 2010 digitalSTROM.org, Schlieren, Switzerland

 digitalSTROM ds485-client library is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 digitalSTROM ds485-client library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with digitalSTROM ds485-client library. If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <getopt.h>
#include <termios.h> /* POSIX terminal control definitions */
#include <pthread.h>
#include <fcntl.h>   /* File control definitions */
#include <sys/unistd.h>
#include <stdlib.h>
#include <stdbool.h>

#include "ds485-client-interface.h"

static int verbose = 1;
static int mainActive = 1;
static int serialLoopback;
static int rxFilterFlags;
static unsigned char ignore_messages[80];

struct message_names_t {
    const char *name;
    int id;
} message_names[] = {
    { "ZONE_COUNT", 0x00 },
    { "ZONE_INFO", 0x01 },
    { "ZONE_MODIFY", 0x02 },
    { "ZONE_DEVICE_COUNT", 0x03 },
    { "ZONE_PROPERTIES", 0x04 },
    { "ZONE_GROUP_MODIFY", 0x10 },
    { "ZONE_GROUP_INFO", 0x11 },
    { "ZONE_GROUP_PROPERTIES", 0x12 },
    { "ZONE_GROUP_SCENE_INFO", 0x18 },
    { "ZONE_GROUP_SCENE_PROPERTIES", 0x19 },
    { "ZONE_GROUP_ACTION_REQUEST", 0x50 },
    { "ZONE_SENSOR_PUSH", 0x56 },
    { "DEVICE_INFO", 0x20 },
    { "DEVICE_STATE", 0x21 },
    { "DEVICE_STATISTICS_REQUEST", 0x22 },
    { "DEVICE_PROPERTIES", 0x23 },
    { "DEVICE_BUTTON_INFO", 0x24 },
    { "DEVICE_BINARY_INPUT", 0x26 },
    { "DEVICE_GROUP_MEMBERSHIP_MODIFY", 0x25 },
    { "TEST_TRANSMISSION_QUALITY", 0x41 },
    { "TEST_DEVICE_PRESENCE", 0x42 },
    { "DEVICE_ACTION_REQUEST", 0x51 },
    { "DEVICE_CONFIG", 0x53 },
    { "DEVICE_SENSOR", 0x52 },
    { "CIRCUIT_REMOVE_INACTIVE_DEVICES", 0x31 },
    { "CIRCUIT_REMOVE_DEVICE", 0x35 },
    { "EVENT_TEST_TRANSMISSION_QUALITY", 0x75 },
    { "EVENT_TEST_DEVICE_PRESENCE", 0x7f },
    { "EVENT_DEVICE_ACCESSIBILITY", 0x71 },
    { "EVENT_DEVICE_STATISTICS", 0x72 },
    { "EVENT_DEVICE_ACTION", 0x73 },
    { "EVENT_DEVICE_LOCAL_ACTION", 0x76 },
    { "EVENT_DEVICE_CONFIG", 0x74 },
    { "EVENT_DEVICE_SENSOR", 0x7d },
    { "EVENT_DEVICE_GIVE_UP", 0x78 },
    { "EVENT_DEVICE_MODEL_CHANGED", 0x79 },
    { "CIRCUIT_PING_DEVICES", 0x30 },
    { "CIRCUIT_ENERGY_METER_VALUE", 0x34 },
    { "CIRCUIT_CURRENT_METER_VALUE", 0x36 },
    { "DSM_INFO", 0x90 },
    { "DSM_PROPERTIES", 0x91 },
    { "DSM_STATE", 0x92 },
    { "DSM_CONFIG", 0xf0 },
    { "DSM_STATISTICS", 0xf2 },
    { "DSM_REBOOT", 0xf4 },
    { "DSM_FACTORY_RESET", 0xf5 },
    { "IMAGE_MAINTENANCE", 0xfe },
    { "IMAGE_TRANSFER", 0xff },
    { "EVENT_DSM_STATE_CHANGE", 0x80 },
    { "ZONE_GROUP_DSLINK_SEND", 0x54 },
    { "DEVICE_DSLINK_SEND", 0x55 },
    { "EVENT_DSLINK", 0x7b },
    { "DSM_DEBUG_MODE", 0xf1 },
    { "EVENT_COMMUNICATION_LOG", 0x77 },
    { "EVENT_DSP_DEBUG_DATA", 0x7e },
    { NULL, 0 }
};

const char *name_of(int msgId)
{
    struct message_names_t *cur = message_names;
    while (cur->name) {
        if (cur->id == msgId) {
            return cur->name;
        }
        cur++;
    }
    return "UNKOWN";
}

void append_ignore_message(unsigned char id)
{
    unsigned i;
    for (i = 0; i < sizeof(ignore_messages); i++) {
        if (!ignore_messages[i]) {
            break;
        }
    }

    if (i >=  sizeof(ignore_messages)) {
        printf ("exceeded size of filter list\n");
        return;
    }

    printf("ignoring %x / %s messages\n", id, name_of(id));
    ignore_messages[i] = id;
}

int ignore_message(unsigned char id)
{
    unsigned i;
    for (i = 0; i < sizeof(ignore_messages); i++) {
        if (ignore_messages[i] == id) {
            return true;
        }
    }
    return false;

}

void formatContainer(char *pTarget, const ds485_container_t *pMessage)
{
    char *pChar = pTarget;
    unsigned char n;

    pChar += sprintf(pChar, "  Type :   %.1x  ", pMessage->containerType);
    pChar += sprintf(pChar, "   Msg :   %.2x/%s\n", pMessage->data[0],
                     name_of(pMessage->data[0]));
    pChar += sprintf(pChar, "  from :   ");
    for (n = 0; n < 12; n++)
        pChar += sprintf(pChar, "%02x", pMessage->sourceId.id[n]);
    pChar += sprintf(pChar, "\n");
    pChar += sprintf(pChar, "    to :   ");
    for (n = 0; n < 12; n++)
        pChar += sprintf(pChar, "%02x", pMessage->destinationId.id[n]);
    if (verbose < 2)
        return;
    pChar += sprintf(pChar, "\n");
    pChar += sprintf(pChar, "param1 : %.2x%.2x ", pMessage->data[2], pMessage->data[1]);
    pChar += sprintf(pChar, "param2 : %.2x%.2x ", pMessage->data[4], pMessage->data[3]);
    pChar += sprintf(pChar, "param3 : %.2x%.2x ", pMessage->data[6], pMessage->data[5]);
    pChar += sprintf(pChar, "param4 : %.2x%.2x\r\n", pMessage->data[8], pMessage->data[7]);
    pChar += sprintf(pChar, "param5 : %.2x%.2x ", pMessage->data[10], pMessage->data[9]);
    pChar += sprintf(pChar, "param6 : %.2x%.2x ", pMessage->data[12], pMessage->data[11]);
    pChar += sprintf(pChar, "param7 : %.2x%.2x ", pMessage->data[14], pMessage->data[13]);
    pChar += sprintf(pChar, "length : %d\r\n", pMessage->length);
}

struct message_t {
    unsigned char id;
    unsigned char remains[DS485_MAX_PAYLOAD_SIZE - 1];
};

int client_callback_state(void *data, const bus_state_t state)
{
    (void)data;
    printf("callback state: %d\n", state);
    return 0;
}
int client_callback_container(void *data, const ds485_container_t *container)
{
    (void)data;
    char text[1000];
    struct message_t *msg = (void *)container->data;
    if (ignore_message(msg->id))
        return 0;
    formatContainer(text, container);
    printf("callback container: %s\n", text);
    return 0;
}

/* Command line parsing */
extern char *optarg;
extern int optind;

/* Serial Device Name of the device we manage */
static char* connspec;
static const char conndefault[] = "serial:/dev/ttyUSB0";

int main(int argc, char *argv[])
{
    int result, flags = 0;
    char c;
    ds485ClientHandle_t handle;
    struct ds485c_callbacks cb;

    connspec = (char *) conndefault;

    /* Look for options */
    while ((c = getopt(argc, argv, "hVv:f:b:Lpi:")) != -1) {
        switch (c)
        {
        case 'v':
            verbose = atoi(optarg);
            printf("Set debug level to: 0x%02x\n", verbose);
            break;
        case 'f':
            rxFilterFlags = strtoul(optarg, NULL, 0);
            printf("Set rx filter to: 0x%02x\n", rxFilterFlags);
            break;
        case 'i':
            append_ignore_message(strtoul(optarg, NULL, 0));
            break;
        case 'L':
            serialLoopback = 1;
            printf("Enable loopback transmission\n");
            break;
        case 'p':
            printf("Enable promiscious mode\n");
            flags = PROMISCUOUS_MODE;
            break;
        default:
            exit(-1);
        }
    }

    /* Last arg is device or connection name. */
    if (optind < argc)
        connspec = strdup(argv[argc - 1]);

    memset(&cb, 0, sizeof(cb));
    cb.container_pkt_cb = client_callback_container;
    cb.state_change_cb = client_callback_state;

    handle = ds485_client_open2(connspec, flags, &cb);
    if (handle == 0) {
        printf("ds485_client_open failed\n");
        return -1;
    }

    ds485_container_t pFrame;
    pFrame.containerType = DS485_CONTAINER_REQUEST;
    pFrame.data[0] = 0x90;
    pFrame.length = 1;
    SetBroadcastDsuid(pFrame.destinationId);
    ds485_client_send_command(handle, &pFrame);

    ds485_container_t pFrame2;
    pFrame2.containerType = DS485_CONTAINER_REQUEST;
    pFrame2.data[0] = 0x91;
    pFrame2.length = 1;
    SetBroadcastDsuid(pFrame2.destinationId);
    ds485_client_send_command(handle, &pFrame2);

#if 0
    {
        uint16_t ZoneId;
        uint8_t NumberOfGroups;
        uint8_t Name[20];
        result = ZoneInfoOfIndex_Get(dsid, 0, &ZoneId, &NumberOfGroups, Name);
        printf("ZoneId: %d\nResult: %x\n", ZoneId, result);
        uint16_t DeviceCount;
        result = ZoneDeviceCount_OnlyActive(dsid, ZoneId, &DeviceCount);
        printf("DeviceCount: %d\nResult: %x\n", DeviceCount, result);
    }

    {
        int i;
        uint32_t RevHW;
        uint32_t RevSW;
        dsid_t DSID;
        uint8_t Name[20];
        result = MeterInfo_Get(dsid, &RevHW, &RevSW, &DSID, Name);

        printf("HW Revision: 0x%x\nSW Revision: 0x%x\ndSID: ", RevHW, RevSW);
        for (i = 0; i < 12; ++i)
        {
            printf("%02x ", DSID[i]);
        }
        printf("\nName: %s\nResult: %x\n", Name, result);
    }
#endif

    while (mainActive) {
        int key = getchar();
        switch (key)
        {
        case '?':
        case 'h':
            break;
        case 'q':
            mainActive = 0;
            break;
        }
    }

    result = ds485_client_close(handle);
    if (result < 0) {
        printf("ds485_client_close failed [%d]\n", result);
        return -1;
    }

    return 0;
}
